﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ProgramovaniSoutez
{
    // třída na logiku hry
    public class Hra
    {
        public int hrac1 = 0;
        public int hrac2 = 0;
        
        Stopwatch timeLimit = new Stopwatch();

        bool timer = false;

        public TimeSpan timespan = TimeSpan.MinValue;
        int ticks = 0;

        Gem[,] pole = new Gem[8, 8];

        Gem[,] helper = null;

        int gemWidth = 50;
        int gemHeight = 50;

        Gem selectedGem = null;
        Gem selectedGem2 = null;

        bool end = false;

        public event EventHandler<string> Ended;
        public event EventHandler<int> ScoreUpdate;
        public event EventHandler<string> TimeUpdate;

        public Hra(bool limit, DateTime time)
        {
            timer = limit;

            if (timer)
            {
                timespan = TimeSpan.FromSeconds(60 * time.Minute + time.Second);
                ticks = (int)timespan.TotalSeconds;
            }
        }

        public PointF Scale { get; set; } = new PointF(1.0f, 1.0f);

        public void Load(string path)
        {
            StreamReader reader = new StreamReader(path);

            string line = string.Empty;

            int i = 0;

            while((line = reader.ReadLine()) != null)
            {
                if(i < 8)
                {
                    for (int j = 0; j < 8; j++)
                        pole[j, i] = new Gem(Gem.Barvy[Int32.Parse(line[j].ToString()) - 1], new Point(j * gemWidth, i * gemHeight));
                }
                else
                {
                    if (i == 8)
                        helper = new Gem[8, line.Length];

                    for(int j = 0; j < line.Length; j++)
                        helper[i - 8, j] = new Gem(Gem.Barvy[Int32.Parse(line[j].ToString()) - 1], new Point((i - 8) * gemWidth, -gemHeight));
                }

                i++;
            }

            if(timer)
                timeLimit.Start();
        }

        public void Draw(Graphics graphics)
        {
            graphics.ScaleTransform(Scale.X, Scale.Y);

            for(int x = 0; x < (helper?.GetLength(0) ?? 0); x++)
            {
                Gem gem = helper[x, 0];

                graphics.FillRectangle(new SolidBrush(gem.Barva), x * gemWidth, 0, gemWidth, gemHeight);
            }

            graphics.TranslateTransform(0, gemHeight);

            for(int x = 0; x < pole.GetLength(0); x++)
            {
                for(int y = 0; y < pole.GetLength(1); y++)
                {
                    Gem gem = pole[x, y];

                    if (gem == null)
                        continue;

                    graphics.FillRectangle(new SolidBrush(gem.Barva), gem.Pozice.X, gem.Pozice.Y, gemWidth, gemHeight);

                    if (selectedGem == gem)
                        graphics.DrawRectangle(new Pen(Brushes.Black, 5.0f), gem.Pozice.X, gem.Pozice.Y, gemWidth, gemHeight);
                }
            }

            graphics.TranslateTransform(0, -gemHeight);

            graphics.DrawLine(new Pen(Brushes.Black, 5.0f), new Point(0, gemHeight), new Point(8 * gemWidth, gemHeight));
        }

        public void Update()
        {
            for (int x = 0; x < pole.GetLength(0); x++)
            {
                for (int y = 0; y < pole.GetLength(1); y++)
                {
                    Gem gem = pole[x, y];

                    if (gem == null)
                        continue;

                    if (gem.Pozice != gem.NovaPozice)
                    {
                        if (gem.NovaPozice.Y > gem.Pozice.Y)
                            gem.Pozice = new Point(gem.Pozice.X, gem.Pozice.Y + 5);

                        if (gem.NovaPozice.Y < gem.Pozice.Y)
                            gem.Pozice = new Point(gem.Pozice.X, gem.Pozice.Y - 5);

                        if (gem.NovaPozice.X > gem.Pozice.X)
                            gem.Pozice = new Point(gem.Pozice.X + 5, gem.Pozice.Y);

                        if (gem.NovaPozice.X < gem.Pozice.X)
                            gem.Pozice = new Point(gem.Pozice.X - 5, gem.Pozice.Y);
                    }
                }
            }

            if (end)
                return;

            if(selectedGem != null && selectedGem2 != null)
            {
                // checking for neighbours
                Point pos1 = getPosition(selectedGem);
                Point pos2 = getPosition(selectedGem2);

                int diffX = pos1.X - pos2.X;
                int diffY = pos1.Y - pos2.Y;

                if(diffX <= 1 && diffX >= -1 && diffY <= 1 && diffY >= -1)
                {
                    bool check = false;

                    if (diffX == 0)
                        check = true;

                    if (diffY == 0)
                        check = true;

                    if(check && (checkPosition(selectedGem, pos2) || checkPosition(selectedGem2, pos1)))
                    {
                        Point pos = selectedGem.NovaPozice;

                        Gem gem1 = selectedGem;
                        Gem gem2 = selectedGem2;

                        selectedGem.NovaPozice = selectedGem2.Pozice;
                        selectedGem2.NovaPozice = pos;

                        pole[pos1.X, pos1.Y] = gem2;
                        pole[pos2.X, pos2.Y] = gem1;
                    }
                }

                selectedGem = null;
                selectedGem2 = null;
            }

            // 
            gemsValues();

            // find and remove
            hrac1 += gemsRemove();

            // move upper gems
            gemsMove();

            // no more moves => end

            if (timer && timeLimit.Elapsed >= timespan)
            {
                end = true;
                
                Ended?.Invoke(this, "Vypršel časový limit");
            }
        }

        public void Update2()
        {
            if(timer && timeLimit.Elapsed < timespan)
            {
                ticks--;

                TimeUpdate?.Invoke(this, $"{ticks / 60}:{ticks - ticks / 60}");
            }
        }

        public void MouseClick(Point pos)
        {
            Point position = new Point((int)((float)pos.X / Scale.X), (int)((float)(pos.Y - gemHeight) / Scale.Y));

            Point location = new Point(position.X / gemWidth, position.Y / gemHeight);

            if(location.X >= 0 && location.X < pole.GetLength(0)
            && location.Y >= 0 && location.Y < pole.GetLength(1))
            {
                if (selectedGem2 == null && selectedGem != null && pole[location.X, location.Y] != selectedGem)
                    selectedGem2 = pole[location.X, location.Y];

                if (selectedGem == null)
                    selectedGem = pole[location.X, location.Y];
            }
        }

        Gem[,] shiftLeft(Gem[,] array, int x)
        {
            Gem[,] demo = new Gem[array.GetLength(0), array.GetLength(1)];

            for (int i = 0; i < array.GetLength(0); i++)
            {
                for (int y = 0; y < array.GetLength(1); y++)
                    demo[i, y] = array[i, y];
            }

            for (int i = 0; i < array.GetLength(1) - 1; i++)
                demo[x, i] = array[x, i + 1];

            demo[x, demo.GetLength(1) - 1] = array[x, 0];

            return demo;
        }

        Point getPosition(Gem gem)
        {
            for (int x = 0; x < pole.GetLength(0); x++)
            {
                for (int y = 0; y < pole.GetLength(1); y++)
                {
                    if (gem == pole[x, y])
                        return new Point(x, y);
                }
            }

            return new Point();
        }

        bool checkPosition(Gem gem, Point targetLoc)
        {
            // 1
            if(targetLoc.Y - 2 >= 0)
            {
                if(pole[targetLoc.X, targetLoc.Y - 1]?.Barva == gem.Barva && pole[targetLoc.X, targetLoc.Y - 1] != gem
                && pole[targetLoc.X, targetLoc.Y - 2]?.Barva == gem.Barva && pole[targetLoc.X, targetLoc.Y - 2] != gem)
                    return true;
            }

            // 2
            if(targetLoc.Y - 1 >= 0 && targetLoc.Y + 1 < pole.GetLength(1))
            {
                if(pole[targetLoc.X, targetLoc.Y - 1]?.Barva == gem.Barva && pole[targetLoc.X, targetLoc.Y - 1] != gem
                && pole[targetLoc.X, targetLoc.Y + 1]?.Barva == gem.Barva && pole[targetLoc.X, targetLoc.Y + 1] != gem)
                    return true;
            }

            // 3
            if (targetLoc.Y + 2 < pole.GetLength(1))
            {
                if(pole[targetLoc.X, targetLoc.Y + 1]?.Barva == gem.Barva && pole[targetLoc.X, targetLoc.Y + 1] != gem
                && pole[targetLoc.X, targetLoc.Y + 2]?.Barva == gem.Barva && pole[targetLoc.X, targetLoc.Y + 2] != gem)
                    return true;
            }

            // 4
            if(targetLoc.X + 2 < pole.GetLength(0))
            {
                if(pole[targetLoc.X + 1, targetLoc.Y]?.Barva == gem.Barva && pole[targetLoc.X + 1, targetLoc.Y] != gem
                && pole[targetLoc.X + 2, targetLoc.Y]?.Barva == gem.Barva && pole[targetLoc.X + 2, targetLoc.Y] != gem)
                    return true;
            }

            // 5
            if (targetLoc.X - 1 >= 0 && targetLoc.X + 1 < pole.GetLength(0))
            {
                if(pole[targetLoc.X - 1, targetLoc.Y]?.Barva == gem.Barva && pole[targetLoc.X - 1, targetLoc.Y] != gem
                && pole[targetLoc.X + 1, targetLoc.Y]?.Barva == gem.Barva && pole[targetLoc.X + 1, targetLoc.Y] != gem)
                    return true;
            }

            // 6
            if (targetLoc.X - 2 >= 0)
            {
                if(pole[targetLoc.X - 1, targetLoc.Y]?.Barva == gem.Barva && pole[targetLoc.X - 1, targetLoc.Y] != gem
                && pole[targetLoc.X - 2, targetLoc.Y]?.Barva == gem.Barva && pole[targetLoc.X - 2, targetLoc.Y] != gem)
                    return true;
            }

            return false;
        }

        void gemsValues()
        {
            for (int x = 0; x < pole.GetLength(0); x++)
            {
                for (int y = 0; y < pole.GetLength(1); y++)
                {
                    Gem gem = pole[x, y];

                    if (gem == null)
                        continue;

                    if(x + 2 < pole.GetLength(0))
                    {
                        if(pole[x + 1, y] != null && pole[x + 2, y] != null
                        && pole[x + 1, y].Barva == gem.Barva
                        && pole[x + 2, y].Barva == gem.Barva)
                        {
                            gem.Value++;
                            pole[x + 1, y].Value++;
                            pole[x + 2, y].Value++;
                        }
                    }

                    if(y + 2 < pole.GetLength(1))
                    {
                        if (pole[x, y + 1] != null && pole[x, y + 2] != null
                        && pole[x, y + 1].Barva == gem.Barva
                        && pole[x, y + 2].Barva == gem.Barva)
                        {
                            gem.Value++;
                            pole[x, y + 1].Value++;
                            pole[x, y + 2].Value++;
                        }
                    }
                }
            }
        }

        // odstanění gemů
        int gemsRemove()
        {
            int score = 0;

            for (int x = 0; x < pole.GetLength(0); x++)
            {
                for (int y = 0; y < pole.GetLength(1); y++)
                {
                    Gem gem = pole[x, y];

                    if (gem == null)
                        continue;

                    if (gem.Value > 0)
                    {
                        score += gem.Value;

                        pole[x, y] = null;
                    }
                }
            }

            ScoreUpdate?.Invoke(this, score);

            return score;
        }

        // posun gemů, blbé je že mezitím co se animují proběhne několikrát update, takže se provede odstranění dalšího sledu 
        void gemsMove()
        {
            for (int x = 0; x < pole.GetLength(0); x++)
            {
                for (int y = pole.GetLength(1) - 1; y >= 0; y--)
                {
                    Gem gem = pole[x, y];

                    if(y == 0 && gem == null)
                    {
                        pole[x, y] = helper[x, 0];
                        pole[x, y].NovaPozice = new Point(gemWidth * x, 0);

                        Gem[,] demo = shiftLeft(helper, x);
                        helper = demo;
                    }
                    
                    if(y > 0)
                    {
                        Gem upper = pole[x, y - 1];

                        if(gem == null && upper != null)
                        {
                            pole[x, y] = upper;
                            pole[x, y - 1] = null;

                            pole[x, y].NovaPozice = new Point(gemWidth * x, gemHeight * y);
                        }
                    }
                }
            }
        }
    }
}
